@@ -96,7 +96,8 @@ Keys in variables list:
     t/r/s/c - show type/repr/str/custom for this variable
     h - toggle highlighting
     @ - toggle repetition at top
-    * - toggle private members
+    * - cycle attribute visibility: public/_private/__dunder__
+    m - toggle method visibility
     w - toggle line wrapping
     n/insert - add new watch expression
     enter - edit options (also to delete)
@@ -144,9 +145,9 @@ OTHER DEALINGS IN THE SOFTWARE.
 # {{{ debugger interface
 
 class Debugger(bdb.Bdb):
-    def __init__(self, steal_output=False):
+    def __init__(self, stdin=None, stdout=None, term_size=None, steal_output=False):
         bdb.Bdb.__init__(self)
-        self.ui = DebuggerUI(self)
+        self.ui = DebuggerUI(self, stdin=stdin, stdout=stdout, term_size=term_size)
         self.steal_output = steal_output
 
         self.setup_state()
@@ -164,24 +165,68 @@ class Debugger(bdb.Bdb):
         for bpoint_descr in load_breakpoints():
             self.set_break(*bpoint_descr)
 
-    def set_trace(self, frame=None):
+    # These (dispatch_line and set_continue) are copied from bdb with the
+    # patch from https://bugs.python.org/issue16482 applied. See
+    # https://github.com/inducer/pudb/pull/90.
+    def dispatch_line(self, frame):
+        if self.stop_here(frame) or self.break_here(frame):
+            self.user_line(frame)
+            if self.quitting:
+                raise bdb.BdbQuit
+            # Do not re-install the local trace when we are finished debugging,
+            # see issues 16482 and 7238.
+            if not sys.gettrace():
+                return None
+        return self.trace_dispatch
+
+    def set_continue(self):
+        # Don't stop except at breakpoints or when finished
+        self._set_stopinfo(self.botframe, None, -1)
+        if not self.breaks:
+            # no breakpoints; run without debugger overhead
+            sys.settrace(None)
+            frame = sys._getframe().f_back
+            while frame:
+                del frame.f_trace
+                if frame is self.botframe:
+                    break
+                frame = frame.f_back
+
+    def set_trace(self, frame=None, as_breakpoint=True):
         """Start debugging from `frame`.
 
         If frame is not specified, debugging starts from caller's frame.
 
-        This is exactly the same as Bdb.set_trace(), sans the self.reset() call.
+        Unlike Bdb.set_trace(), this does not call self.reset(), which causes
+        the debugger to enter bdb source code. This also implements treating
+        set_trace() calls as breakpoints in the PuDB UI.
         """
         if frame is None:
-            frame = sys._getframe().f_back
+            frame = thisframe = sys._getframe().f_back
+        else:
+            thisframe = frame
         # See pudb issue #52. If this works well enough we should upstream to
         # stdlib bdb.py.
         #self.reset()
+
         while frame:
             frame.f_trace = self.trace_dispatch
             self.botframe = frame
             frame = frame.f_back
-        self.set_step()
-        sys.settrace(self.trace_dispatch)
+
+        thisframe_info = (
+                self.canonic(thisframe.f_code.co_filename), thisframe.f_lineno)
+        if thisframe_info not in self.set_traces or self.set_traces[thisframe_info]:
+            if as_breakpoint:
+                self.set_traces[thisframe_info] = True
+                if self.ui.source_code_provider is not None:
+                    self.ui.set_source_code_provider(
+                            self.ui.source_code_provider, force_update=True)
+
+            self.set_step()
+            sys.settrace(self.trace_dispatch)
+        else:
+            return
 
     def save_breakpoints(self):
         from pudb.settings import save_breakpoints
@@ -201,6 +246,9 @@ class Debugger(bdb.Bdb):
         self._wait_for_mainpyfile = False
         self.current_bp = None
         self.post_mortem = False
+        # Mapping of (filename, lineno) to bool. If True, will stop on the
+        # set_trace() call at that location.
+        self.set_traces = {}
 
     def restart(self):
         from linecache import checkcache
@@ -389,42 +437,51 @@ except ImportError:
     curses = None
 
 
-want_curses_display = (
-        CONFIG["display"] == "curses"
-        or (
-            CONFIG["display"] == "auto"
-            and
-            not os.environ.get("TERM", "").startswith("xterm")))
-
 from urwid.raw_display import Screen as RawScreen
-if want_curses_display:
-    try:
-        from urwid.curses_display import Screen
-    except ImportError:
-        Screen = RawScreen
-else:
-    Screen = RawScreen
-
-del want_curses_display
+try:
+    from urwid.curses_display import Screen as CursesScreen
+except ImportError:
+    CursesScreen = None
 
 
-class ThreadsafeScreen(Screen):
+class ThreadsafeScreenMixin(object):
     "A Screen subclass that doesn't crash when running from a non-main thread."
 
     def signal_init(self):
         "Initialize signal handler, ignoring errors silently."
         try:
-            super(ThreadsafeScreen, self).signal_init()
+            super(ThreadsafeScreenMixin, self).signal_init()
         except ValueError:
             pass
 
     def signal_restore(self):
         "Restore default signal handler, ignoring errors silently."
         try:
-            super(ThreadsafeScreen, self).signal_restore()
+            super(ThreadsafeScreenMixin, self).signal_restore()
         except ValueError:
             pass
 
+
+class ThreadsafeRawScreen(ThreadsafeScreenMixin, RawScreen):
+    pass
+
+
+class ThreadsafeFixedSizeRawScreen(ThreadsafeScreenMixin, RawScreen):
+    def __init__(self, **kwargs):
+        self._term_size = kwargs.pop("term_size", None)
+        super(ThreadsafeFixedSizeRawScreen, self).__init__(**kwargs)
+
+    def get_cols_rows(self):
+        if self._term_size is not None:
+            return self._term_size
+        else:
+            return 80, 24
+
+
+if curses is not None:
+    class ThreadsafeCursesScreen(ThreadsafeScreenMixin, RawScreen):
+        pass
+
 # }}}
 
 
@@ -486,9 +543,11 @@ class FileSourceCodeProvider(SourceCodeProvider):
         from pudb.source_view import SourceLine, format_source
 
         if self.file_name == "<string>":
-            return [SourceLine(self, self.file_name)]
+            return [SourceLine(debugger_ui, self.file_name)]
 
-        breakpoints = debugger_ui.debugger.get_file_breaks(self.file_name)
+        breakpoints = debugger_ui.debugger.get_file_breaks(self.file_name)[:]
+        breakpoints += [i for f, i in debugger_ui.debugger.set_traces if f
+            == self.file_name and debugger_ui.debugger.set_traces[f, i]]
         try:
             from linecache import getlines
             lines = getlines(self.file_name)
@@ -506,10 +565,10 @@ class FileSourceCodeProvider(SourceCodeProvider):
             return format_source(debugger_ui, decoded_lines, set(breakpoints))
         except:
             from pudb.lowlevel import format_exception
-            self.message("Could not load source file '%s':\n\n%s" % (
+            debugger_ui.message("Could not load source file '%s':\n\n%s" % (
                 self.file_name, "".join(format_exception(sys.exc_info()))),
                 title="Source Code Load Error")
-            return [SourceLine(self,
+            return [SourceLine(debugger_ui,
                 "Error while loading '%s'." % self.file_name)]
 
 
@@ -547,8 +606,6 @@ class DirectSourceCodeProvider(SourceCodeProvider):
         for i, l in enumerate(lines):
             if hasattr(l, "decode"):
                 l = l.decode(source_enc)
-            else:
-                l = l.decode(source_enc)
 
             if i+1 < len(lines):
                 l += "\n"
@@ -563,7 +620,7 @@ class DirectSourceCodeProvider(SourceCodeProvider):
 class DebuggerUI(FrameVarInfoKeeper):
     # {{{ constructor
 
-    def __init__(self, dbg):
+    def __init__(self, dbg, stdin, stdout, term_size):
         FrameVarInfoKeeper.__init__(self)
 
         self.debugger = dbg
@@ -711,9 +768,12 @@ class DebuggerUI(FrameVarInfoKeeper):
             elif key == "@":
                 iinfo.repeated_at_top = not iinfo.repeated_at_top
             elif key == "*":
-                iinfo.show_private_members = not iinfo.show_private_members
+                levels = ["public", "private", "all", "public"]
+                iinfo.access_level = levels[levels.index(iinfo.access_level)+1]
             elif key == "w":
                 iinfo.wrap = not iinfo.wrap
+            elif key == "m":
+                iinfo.show_methods = not iinfo.show_methods
 
             self.update_var_view()
 
@@ -748,32 +808,45 @@ class DebuggerUI(FrameVarInfoKeeper):
 
                 title = "Variable Inspection Options"
 
-            rb_grp = []
-            rb_show_type = urwid.RadioButton(rb_grp, "Show Type",
+            rb_grp_show = []
+            rb_show_type = urwid.RadioButton(rb_grp_show, "Show Type",
                     iinfo.display_type == "type")
-            rb_show_repr = urwid.RadioButton(rb_grp, "Show repr()",
+            rb_show_repr = urwid.RadioButton(rb_grp_show, "Show repr()",
                     iinfo.display_type == "repr")
-            rb_show_str = urwid.RadioButton(rb_grp, "Show str()",
+            rb_show_str = urwid.RadioButton(rb_grp_show, "Show str()",
                     iinfo.display_type == "str")
-            rb_show_custom = urwid.RadioButton(rb_grp, "Show custom (set in prefs)",
+            rb_show_custom = urwid.RadioButton(
+                    rb_grp_show, "Show custom (set in prefs)",
                     iinfo.display_type == CONFIG["custom_stringifier"])
 
+            rb_grp_access = []
+            rb_access_public = urwid.RadioButton(rb_grp_access, "Public members",
+                    iinfo.access_level == "public")
+            rb_access_private = urwid.RadioButton(
+                    rb_grp_access, "Public and private members",
+                    iinfo.access_level == "private")
+            rb_access_all = urwid.RadioButton(
+                    rb_grp_access, "All members (including __dunder__)",
+                    iinfo.access_level == "all")
+
             wrap_checkbox = urwid.CheckBox("Line Wrap", iinfo.wrap)
             expanded_checkbox = urwid.CheckBox("Expanded", iinfo.show_detail)
             highlighted_checkbox = urwid.CheckBox("Highlighted", iinfo.highlighted)
             repeated_at_top_checkbox = urwid.CheckBox(
                     "Repeated at top", iinfo.repeated_at_top)
-            show_private_checkbox = urwid.CheckBox(
-                    "Show private members", iinfo.show_private_members)
+            show_methods_checkbox = urwid.CheckBox(
+                    "Show methods", iinfo.show_methods)
 
             lb = urwid.ListBox(urwid.SimpleListWalker(
-                id_segment+rb_grp+[
-                    urwid.Text(""),
+                id_segment +
+                rb_grp_show + [urwid.Text("")] +
+                rb_grp_access + [urwid.Text("")] +
+                [
                     wrap_checkbox,
                     expanded_checkbox,
                     highlighted_checkbox,
                     repeated_at_top_checkbox,
-                    show_private_checkbox,
+                    show_methods_checkbox,
                 ]))
 
             result = self.dialog(lb, buttons, title=title)
@@ -783,7 +856,7 @@ class DebuggerUI(FrameVarInfoKeeper):
                 iinfo.wrap = wrap_checkbox.get_state()
                 iinfo.highlighted = highlighted_checkbox.get_state()
                 iinfo.repeated_at_top = repeated_at_top_checkbox.get_state()
-                iinfo.show_private_members = show_private_checkbox.get_state()
+                iinfo.show_methods = show_methods_checkbox.get_state()
 
                 if rb_show_type.get_state():
                     iinfo.display_type = "type"
@@ -794,6 +867,13 @@ class DebuggerUI(FrameVarInfoKeeper):
                 elif rb_show_custom.get_state():
                     iinfo.display_type = CONFIG["custom_stringifier"]
 
+                if rb_access_public.get_state():
+                    iinfo.access_level = "public"
+                elif rb_access_private.get_state():
+                    iinfo.access_level = "private"
+                elif rb_access_all.get_state():
+                    iinfo.access_level = "all"
+
                 if var.watch_expr is not None:
                     var.watch_expr.expression = watch_edit.get_edit_text()
 
@@ -833,6 +913,7 @@ class DebuggerUI(FrameVarInfoKeeper):
         self.var_list.listen("@", change_var_state)
         self.var_list.listen("*", change_var_state)
         self.var_list.listen("w", change_var_state)
+        self.var_list.listen("m", change_var_state)
         self.var_list.listen("enter", edit_inspector_detail)
         self.var_list.listen("n", insert_watch)
         self.var_list.listen("insert", insert_watch)
@@ -1127,21 +1208,31 @@ class DebuggerUI(FrameVarInfoKeeper):
                     err = self.debugger.clear_break(bp_source_identifier, lineno)
                     sline.set_breakpoint(False)
                 else:
+                    file_lineno = (bp_source_identifier, lineno)
+                    if file_lineno in self.debugger.set_traces:
+                        self.debugger.set_traces[file_lineno] = \
+                                not self.debugger.set_traces[file_lineno]
+                        sline.set_breakpoint(self.debugger.set_traces[file_lineno])
+                        return
+
                     from pudb.lowlevel import get_breakpoint_invalid_reason
                     invalid_reason = get_breakpoint_invalid_reason(
                             bp_source_identifier, pos+1)
 
                     if invalid_reason is not None:
                         do_set = not self.dialog(
-                                urwid.ListBox(urwid.SimpleListWalker([
-                                    urwid.Text("The breakpoint you just set may be "
-                                        "invalid, for the following reason:\n\n"
-                                        + invalid_reason),
-                                    ])), [
-                                        ("Cancel", True),
-                                        ("Set Anyway", False),
-                                        ], title="Possibly Invalid Breakpoint",
-                                    focus_buttons=True)
+                                urwid.ListBox(
+                                    urwid.SimpleListWalker([
+                                        urwid.Text(
+                                            "The breakpoint you just set may be "
+                                            "invalid, for the following reason:\n\n"
+                                            + invalid_reason),
+                                        ])), [
+                                            ("Cancel", True),
+                                            ("Set Anyway", False),
+                                            ],
+                                title="Possibly Invalid Breakpoint",
+                                focus_buttons=True)
                     else:
                         do_set = True
 
@@ -1187,7 +1278,9 @@ class DebuggerUI(FrameVarInfoKeeper):
 
             def build_filtered_mod_list(filt_string=""):
                 modules = sorted(name
-                        for name, mod in sys.modules.items()
+                        # mod_exists may change the size of sys.modules,
+                        # causing this to crash. Copy to a list.
+                        for name, mod in list(sys.modules.items())
                         if mod_exists(mod))
 
                 result = [urwid.AttrMap(SelectableText(mod),
@@ -1254,7 +1347,7 @@ class DebuggerUI(FrameVarInfoKeeper):
                                     format_exception(sys.exc_info()))),
                                 title="Import Error")
                         else:
-                            show_mod(sys.modules[str(new_mod_name)])
+                            show_mod(__import__(str(new_mod_name)))
                             break
                     else:
                         show_mod(sys.modules[widget.base_widget.get_text()[0]])
@@ -1476,6 +1569,34 @@ class DebuggerUI(FrameVarInfoKeeper):
         def cmdline_history_next(w, size, key):
             cmdline_history_browse(1)
 
+        def cmdline_start_of_line(w, size, key):
+            self.cmdline_edit.edit_pos = 0
+
+        def cmdline_end_of_line(w, size, key):
+            self.cmdline_edit.edit_pos = len(self.cmdline_edit.edit_text)
+
+        def cmdline_del_word(w, size, key):
+            pos = self.cmdline_edit.edit_pos
+            before, after = (
+                    self.cmdline_edit.edit_text[:pos],
+                    self.cmdline_edit.edit_text[pos:])
+            before = before[::-1]
+            before = before.lstrip()
+            i = 0
+            while i < len(before):
+                if not before[i].isspace():
+                    i += 1
+                else:
+                    break
+
+            self.cmdline_edit.edit_text = before[i:][::-1] + after
+            self.cmdline_edit.edit_post = len(before[i:])
+
+        def cmdline_del_to_start_of_line(w, size, key):
+            pos = self.cmdline_edit.edit_pos
+            self.cmdline_edit.edit_text = self.cmdline_edit.edit_text[pos:]
+            self.cmdline_edit.edit_pos = 0
+
         def toggle_cmdline_focus(w, size, key):
             self.columns.set_focus(self.lhs_col)
             if self.lhs_col.get_focus() is self.cmdline_sigwrap:
@@ -1491,6 +1612,10 @@ class DebuggerUI(FrameVarInfoKeeper):
         self.cmdline_edit_sigwrap.listen("ctrl p", cmdline_history_prev)
         self.cmdline_edit_sigwrap.listen("esc", toggle_cmdline_focus)
         self.cmdline_edit_sigwrap.listen("ctrl d", toggle_cmdline_focus)
+        self.cmdline_edit_sigwrap.listen("ctrl a", cmdline_start_of_line)
+        self.cmdline_edit_sigwrap.listen("ctrl e", cmdline_end_of_line)
+        self.cmdline_edit_sigwrap.listen("ctrl w", cmdline_del_word)
+        self.cmdline_edit_sigwrap.listen("ctrl u", cmdline_del_to_start_of_line)
 
         self.top.listen("ctrl x", toggle_cmdline_focus)
 
@@ -1633,6 +1758,8 @@ class DebuggerUI(FrameVarInfoKeeper):
                 runner = shell.run_ipython_shell
             elif shell.HAVE_BPYTHON and CONFIG["shell"] == "bpython":
                 runner = shell.run_bpython_shell
+            elif shell.HAVE_PTPYTHON and CONFIG["shell"] == "ptpython":
+                runner = shell.run_ptpython_shell
             else:
                 runner = shell.run_classic_shell
 
@@ -1657,7 +1784,7 @@ class DebuggerUI(FrameVarInfoKeeper):
             def __init__(self, idx):
                 self.idx = idx
 
-            def __call__(subself, w, size, key):
+            def __call__(subself, w, size, key):  # noqa
                 self.columns.set_focus(self.rhs_col_sigwrap)
                 self.rhs_col.set_focus(self.rhs_col.widget_list[subself.idx])
 
@@ -1694,7 +1821,36 @@ class DebuggerUI(FrameVarInfoKeeper):
 
         # {{{ setup
 
-        self.screen = ThreadsafeScreen()
+        want_curses_display = (
+                CONFIG["display"] == "curses"
+                or (
+                    CONFIG["display"] == "auto"
+                    and
+                    not (
+                        os.environ.get("TERM", "").startswith("xterm")
+                        or
+                        os.environ.get("TERM", "").startswith("rxvt")
+                    )))
+
+        if (want_curses_display
+                and not (stdin is not None or stdout is not None)
+                and CursesScreen is not None):
+            self.screen = ThreadsafeCursesScreen()
+        else:
+            screen_kwargs = {}
+            if stdin is not None:
+                screen_kwargs["input"] = stdin
+            if stdout is not None:
+                screen_kwargs["output"] = stdout
+            if term_size is not None:
+                screen_kwargs["term_size"] = term_size
+
+            if screen_kwargs:
+                self.screen = ThreadsafeFixedSizeRawScreen(**screen_kwargs)
+            else:
+                self.screen = ThreadsafeRawScreen()
+
+        del want_curses_display
 
         if curses:
             try:
@@ -1749,13 +1905,13 @@ class DebuggerUI(FrameVarInfoKeeper):
             title=None, bind_enter_esc=True, focus_buttons=False,
             extra_bindings=[]):
         class ResultSetter:
-            def __init__(subself, res):
+            def __init__(subself, res):  # noqa
                 subself.res = res
 
-            def __call__(subself, btn):
+            def __call__(subself, btn):  # noqa
                 self.quit_event_loop = [subself.res]
 
-        Attr = urwid.AttrMap
+        Attr = urwid.AttrMap  # noqa
 
         if bind_enter_esc:
             content = SignalWrap(content)
@@ -1796,10 +1952,10 @@ class DebuggerUI(FrameVarInfoKeeper):
                 w])
 
         class ResultSetter:
-            def __init__(subself, res):
+            def __init__(subself, res):  # noqa
                 subself.res = res
 
-            def __call__(subself, w, size, key):
+            def __call__(subself, w, size, key):  # noqa
                 self.quit_event_loop = [subself.res]
 
         w = SignalWrap(w)
@@ -1919,7 +2075,7 @@ class DebuggerUI(FrameVarInfoKeeper):
                 self.message("Package 'pygments' not found. "
                         "Syntax highlighting disabled.")
 
-        WELCOME_LEVEL = "e022"
+        WELCOME_LEVEL = "e026"  # noqa
         if CONFIG["seen_welcome"] < WELCOME_LEVEL:
             CONFIG["seen_welcome"] = WELCOME_LEVEL
             from pudb import VERSION
@@ -1936,10 +2092,34 @@ class DebuggerUI(FrameVarInfoKeeper):
                     "(invoked by hitting '?' after this message) should get you "
                     "on your way.\n"
 
+                    "\nChanges in version 2016.1:\n\n"
+                    "- Fix module browser on Py3.\n"
+
+                    "\nChanges in version 2015.4:\n\n"
+                    "- Support for (somewhat rudimentary) remote debugging\n"
+                    "  through a telnet connection.\n"
+                    "- Fix debugging of generated code in Python 3.\n"
+
+                    "\nChanges in version 2015.3:\n\n"
+                    "- Disable set_trace lines from the UI (Aaron Meurer)\n"
+                    "- Better control over attribute visibility (Ned Batchelder)\n"
+
+                    "\nChanges in version 2015.2:\n\n"
+                    "- ptpython support (P. Varet)\n"
+                    "- Improved rxvt support (Louper Rouch)\n"
+                    "- More keyboard shortcuts in the command line"
+                    "(Alex Sheluchin)\n"
+
+                    "\nChanges in version 2015.1:\n\n"
+                    "- Add solarized theme (Rinat Shigapov)\n"
+                    "- More keyboard shortcuts in the command line"
+                    "(Alexander Corwin)\n"
+
                     "\nChanges in version 2014.1:\n\n"
                     "- Make prompt-on-quit optional (Mike Burr)\n"
                     "- Make tab completion in the built-in shell saner\n"
-                    "- Fix handling of unicode source\n  (reported by Morten Nielsen and Buck Golemon)\n"
+                    "- Fix handling of unicode source\n"
+                    "  (reported by Morten Nielsen and Buck Golemon)\n"
 
                     "\nChanges in version 2013.5.1:\n\n"
                     "- Fix loading of saved breakpoint conditions "
